"""Reads the icpp.toml file"""

import sys
import glob
from pathlib import Path
from typing import Any, Dict, List, Set
import typer
from icpp import config_default

# `tomllib` was introduced in python 3.11
# for earlier versions, use `tomli` instead
if sys.version_info >= (3, 11):
    import tomllib
else:
    try:
        import tomli as tomllib
    except ImportError:
        typer.echo("ERROR: cannot find python module tomli")
        sys.exit(1)

# Look for `icpp.toml`in the current folder
icpp_toml_path = config_default.ICPP_TOML_PATH
if not icpp_toml_path.exists():
    typer.echo(
        f"ERROR: There is no {config_default.ICPP_TOML_PATH} file in the "
        f"current working directory: \n"
        f"        {icpp_toml_path.parent.resolve()} \n"
        f"       Maybe you're in the wrong folder, \n"
        f"       or else can you please create one, so I know what to build?"
    )
    sys.exit(1)


def validate(d_in: Dict[Any, Any]) -> None:
    """Validates if required fields are present"""
    if "build-wasm" not in d_in.keys():
        typer.echo(
            f"ERROR: Missing table '[build-wasm]'\n"
            f"       {icpp_toml_path.resolve()}"
        )
        sys.exit(1)

    dwasm = d_in.get("build-wasm", {})
    if "canister" not in dwasm.keys():
        typer.echo(
            f"ERROR: Missing key 'canister' in '[build-wasm]' table.\n"
            f"       {icpp_toml_path.resolve()}"
        )
        sys.exit(1)

    if "did_path" not in dwasm.keys():
        typer.echo(
            f"ERROR: Missing key 'did_path' in '[build-wasm]' table.\n"
            f"       {icpp_toml_path.resolve()}"
        )
        sys.exit(1)


def read_build_wasm_table(d_in: Dict[Any, Any]) -> Dict[Any, Any]:
    """Reads and processes the '[build-wasm]' table."""
    d = {}
    d["canister"] = d_in.get("canister", "default")
    d["did_path"] = icpp_toml_path.parent / Path(d_in.get("did_path", "")).resolve()
    d["did_file"] = str(d["did_path"]) + " "

    icpp_toml_path_dir = icpp_toml_path.parent.resolve()
    d["cpp_include_dirs"] = [icpp_toml_path_dir, icpp_toml_path_dir / "src"]
    d["c_include_dirs"] = [icpp_toml_path_dir, icpp_toml_path_dir / "src"]
    read_build_table_common(d, d_in)

    return d


def read_libraries(
    d_in: List[Dict[Any, Any]],
    build_wasm_: Dict[Any, Any],
    build_native_: Dict[Any, Any],
    native: bool = False,
) -> List[Dict[Any, Any]]:
    """Reads and processes the '[[build-library]]' tables."""
    libs = []
    seen_lib_names = set()

    # First add the libraries from icpp.toml
    for library_in in d_in:
        library = {}

        library["lib_name"] = library_in.get("lib_name", "")
        if library["lib_name"] in seen_lib_names:
            typer.echo(
                f"ERROR: duplicate 'lib_name' found: {library['lib_name']}\n"
                f"       {icpp_toml_path.resolve()}"
            )
            sys.exit(1)
        seen_lib_names.add(library["lib_name"])

        read_build_table_common(library, library_in)
        libs.append(library)

    # If not defined in the icpp.toml, add the library for the IC & CANDID files
    if "__ic_candid__" not in seen_lib_names:
        library = library_ic_candid(build_wasm_, build_native_, native)
        seen_lib_names.add(library["lib_name"])
        libs.append(library)

    return libs


def read_build_native_table(d_in: Dict[Any, Any]) -> Dict[Any, Any]:
    """Reads and processes the '[build-native]' table."""
    d: Dict[Any, Any] = {}
    icpp_toml_path_dir = icpp_toml_path.parent.resolve()
    d["cpp_include_dirs"] = [icpp_toml_path_dir / "native"]
    d["c_include_dirs"] = [icpp_toml_path_dir, icpp_toml_path_dir / "native"]
    read_build_table_common(d, d_in)
    return d


def library_ic_candid(
    build_wasm_: Dict[Any, Any], build_native_: Dict[Any, Any], native: bool = False
) -> Dict[Any, Any]:
    """Creates a dict for a static library contain the IC and CANDID files"""

    d: Dict[Any, Any] = {}

    if native:
        d = build_native_.copy()
        d["cpp_files_list"] = config_default.MOCKIC_CPP_FILES_LIST
        d["c_files_list"] = config_default.MOCKIC_C_FILES_LIST
        d["cpp_files"] = config_default.MOCKIC_CPP_FILES
        d["cpp_header_files"] = config_default.MOCKIC_HEADER_FILES
        d["c_files"] = config_default.MOCKIC_C_FILES
    else:
        d = build_wasm_.copy()
        d["cpp_files_list"] = config_default.IC_CPP_FILES_LIST
        d["c_files_list"] = config_default.IC_C_FILES_LIST
        d["cpp_files"] = config_default.IC_CPP_FILES
        d["cpp_header_files"] = config_default.IC_HEADER_FILES
        d["c_files"] = config_default.IC_C_FILES

    d["lib_name"] = "__ic_candid__"

    # if non-empty, these defaults will overwrite internal settings
    # d["cpp_compile_flags_defaults"] = []
    # d["c_compile_flags_defaults"] = []

    #
    # all in one compiler uses a long string of strings, not a list
    #
    d["cpp_include_flags"] = ""
    # d["cpp_compile_flags_s"] = ""
    d["c_include_flags"] = ""
    # d["c_header_files"] = ""
    # d["c_compile_flags_s"] = ""

    #
    # overwriting defaults also uses strings
    #
    # d["cpp_compile_flags_defaults_s"] = ""
    # d["c_compile_flags_defaults_s"] = ""

    # d["overwrite_default_CFLAGS"] = False
    # d["overwrite_default_CPPFLAGS"] = False
    # d["overwrite_default_LDFLAGS"] = False

    return d


def read_build_table_common(d: Dict[Any, Any], d_in: Dict[Any, Any]) -> None:
    """Reads the common fields of '[build-wasm]' & '[build-native]'"""

    # Helper function to expand wildcards
    def expand_paths(patterns: List[str]) -> List[Path]:
        paths: Set[Path] = set()  # Using a set to store unique paths
        for pattern in patterns:
            absolute_pattern = icpp_toml_path.parent.resolve() / pattern
            paths.update(map(Path, glob.glob(str(absolute_pattern))))
        return list(paths)

    d["cpp_paths"] = expand_paths(d_in.get("cpp_paths", []))
    if "cpp_include_dirs" in d.keys():
        d["cpp_include_dirs"].extend(expand_paths(d_in.get("cpp_include_dirs", [])))
    else:
        d["cpp_include_dirs"] = expand_paths(d_in.get("cpp_include_dirs", []))
    d["cpp_header_paths"] = expand_paths(d_in.get("cpp_header_paths", []))
    d["cpp_compile_flags"] = d_in.get("cpp_compile_flags", [])
    d["cpp_link_flags"] = d_in.get("cpp_link_flags", [])

    d["c_paths"] = expand_paths(d_in.get("c_paths", []))
    if "c_include_dirs" in d.keys():
        d["c_include_dirs"].extend(expand_paths(d_in.get("c_include_dirs", [])))
    else:
        d["c_include_dirs"] = expand_paths(d_in.get("c_include_dirs", []))

    d["c_header_paths"] = expand_paths(d_in.get("c_header_paths", []))
    d["c_compile_flags"] = d_in.get("c_compile_flags", [])

    # if non-empty, these defaults will overwrite internal settings
    d["cpp_compile_flags_defaults"] = d_in.get("cpp_compile_flags_defaults", [])
    d["cpp_link_flags_defaults"] = d_in.get("cpp_link_flags_defaults", [])
    d["c_compile_flags_defaults"] = d_in.get("c_compile_flags_defaults", [])

    #
    # concurrent compiler uses a list of strings for the files to compile
    #
    d["cpp_files_list"] = [str(x) for x in d["cpp_paths"]]

    d["c_files_list"] = [str(x) for x in d["c_paths"]]

    #
    # all in one compiler uses a long string of strings, not a list
    #
    d["cpp_files"] = " ".join([str(x) for x in d["cpp_paths"]]) + " "
    d["cpp_include_flags"] = (
        " ".join(["-I " + str(path) for path in d["cpp_include_dirs"]]) + " "
    )
    d["cpp_header_files"] = " ".join([str(x) for x in d["cpp_header_paths"]]) + " "
    d["cpp_compile_flags_s"] = " ".join([str(x) for x in d["cpp_compile_flags"]]) + " "

    d["c_files"] = " ".join([str(x) for x in d["c_paths"]]) + " "
    d["c_include_flags"] = (
        " ".join(["-I " + str(path) for path in d["c_include_dirs"]]) + " "
    )
    d["c_header_files"] = " ".join([str(x) for x in d["c_header_paths"]]) + " "
    d["c_compile_flags_s"] = " ".join([str(x) for x in d["c_compile_flags"]]) + " "

    #
    # linker uses a long string of strings, not a list
    #
    d["cpp_link_flags_s"] = " ".join([str(x) for x in d["cpp_link_flags"]]) + " "

    #
    # overwriting defaults also uses strings
    #
    d["cpp_compile_flags_defaults_s"] = (
        " ".join([str(x) for x in d["cpp_compile_flags_defaults"]]) + " "
    )
    d["c_compile_flags_defaults_s"] = (
        " ".join([str(x) for x in d["c_compile_flags_defaults"]]) + " "
    )
    d["cpp_link_flags_defaults_s"] = (
        " ".join([str(x) for x in d["cpp_link_flags_defaults"]]) + " "
    )

    d["overwrite_default_CFLAGS"] = False
    d["overwrite_default_CPPFLAGS"] = False
    d["overwrite_default_LDFLAGS"] = False
    if len(d["c_compile_flags_defaults"]) > 0:
        d["overwrite_default_CFLAGS"] = True
    if len(d["cpp_compile_flags_defaults"]) > 0:
        d["overwrite_default_CPPFLAGS"] = True
    if len(d["cpp_link_flags_defaults"]) > 0:
        d["overwrite_default_LDFLAGS"] = True


#
# Read the top level icpp.toml
#

with open(icpp_toml_path, "rb") as f:
    data = tomllib.load(f)

validate(data)

build_wasm: Dict[Any, Any] = read_build_wasm_table(data.get("build-wasm", {}))
build_native: Dict[Any, Any] = read_build_native_table(data.get("build-native", {}))
libraries: List[Dict[Any, Any]] = read_libraries(
    data.get("build-library", []), build_wasm, build_native, native=False
)
libraries_native: List[Dict[Any, Any]] = read_libraries(
    data.get("build-library-native", []), build_wasm, build_native, native=True
)



"""
Defines the version for the compatible MinGW-w64 compiler from https://winlibs.com/.

(-) Release scripts might modify this automatically...
(-) do not add anything but the version number of the wasi-sdk compiler here!

"""

__version__ = "13.2.0posix-18.1.5-11.0.1-ucrt-r8"



"""Handles 'icpp install-mingw64'"""

import sys
import platform
import shutil
import math
import zipfile
import requests
import enlighten  # type: ignore
import typer
from icpp.__main__ import app
from icpp import config_default
from icpp import __version_mingw64__

OS_SYSTEM = platform.system()
OS_PROCESSOR = platform.processor()

# https://winlibs.com/
# Always select a version for:
# - Win64 - x86_64
# - UCRT runtime
MINGW64_ZIP = (
    "winlibs-x86_64-posix-seh-gcc-13.2.0-llvm-18.1.5-mingw-w64ucrt-11.0.1-r8.zip"
)
MINGW64_URL = f"https://github.com/brechtsanders/winlibs_mingw/releases/download/{__version_mingw64__}/{MINGW64_ZIP}"  # pylint: disable=line-too-long


@app.command()
def install_mingw64() -> None:
    """Installs MinGW-w64 compiler (Windows only).

    Compiler will be installed in ~/.icpp"""

    if OS_SYSTEM != "Windows":
        typer.echo("ERROR: MinGw-w64 is only for Windows, not for your system.")
        sys.exit(1)

    # ----------------------------------------------------------------
    typer.echo(f"Installing MinGW-w64 compiler into {config_default.MINGW64_ROOT}")

    fpath = config_default.MINGW64_COMPILER_ROOT / MINGW64_ZIP
    fpath.unlink(missing_ok=True)

    try:
        shutil.rmtree(config_default.MINGW64_COMPILER_ROOT)
    except FileNotFoundError:
        pass
    except OSError as e:
        typer.echo(f"Warning: {e.strerror}")

    # ----------------------------------------------------------------

    try:
        num_steps = 3
        nstep = 1
        ################################################################################
        # Download the zip file

        # Create the directories if they do not yet exist
        fpath.parent.mkdir(parents=True, exist_ok=True)

        use_progress_bar = True

        with requests.Session() as s:
            r = s.get(MINGW64_URL, stream=use_progress_bar)
            if r.status_code != 200:
                typer.echo(
                    f"Tried to fetch mingw64 zip file from GitHub, "
                    f"but response status code is "
                    f"{r.status_code}. Please retry later..."
                )
                typer.echo(f"url = {MINGW64_URL}")
                sys.exit(1)

            typer.echo(f"- {nstep}/{num_steps} Downloading mingw64: {MINGW64_URL}")
            # typer.echo(f"Saving zip file as  : {fpath}")

            config_default.MINGW64_ROOT.mkdir(parents=True, exist_ok=True)
            if not use_progress_bar:
                with open(fpath, "wb") as f:
                    f.write(r.content)
            else:
                # https://stackoverflow.com/a/63832993/5480536
                progress_bar_manager = enlighten.get_manager()
                dlen = int(r.headers.get("Content-Length", "0")) or None

                with progress_bar_manager.counter(
                    color="cyan",
                    total=dlen and math.ceil(dlen / 2**20),
                    unit="MiB",
                    leave=False,
                ) as ctr, open(fpath, "wb", buffering=2**24) as f:
                    for chunk in r.iter_content(chunk_size=2**20):
                        # print(chunk[-16:].hex().upper())
                        f.write(chunk)
                        ctr.update()

        ################################################################################
        # Unzip the zip file
        nstep += 1
        typer.echo(f"- {nstep}/{num_steps} Unzipping {fpath}")

        if not use_progress_bar:
            with zipfile.ZipFile(fpath, "r") as zip_ref:
                zip_ref.extractall(path=config_default.MINGW64_COMPILER_ROOT)
        else:
            with zipfile.ZipFile(fpath, "r") as zip_ref:
                # All the files in the zipfile
                file_list = zip_ref.namelist()

                # Create a progress bar
                progress_bar = progress_bar_manager.counter(
                    color="cyan", total=len(file_list), leave=False, unit="files"
                )

                for file in file_list:
                    zip_ref.extract(file, config_default.MINGW64_COMPILER_ROOT)
                    progress_bar.update()

                # Close the progress bar
                progress_bar.close()

            # Now we can delete the progress_bar_manager
            progress_bar_manager.stop()

        # Cleaning up
        nstep += 1
        typer.echo(f"- {nstep}/{num_steps} Cleaning up")
        fpath.unlink(missing_ok=True)

        typer.echo(
            f"\nSuccessfully installed MinGW-w64 in "
            f"{config_default.MINGW64_COMPILER_ROOT }"
        )
        try:
            typer.echo("💯 🎉 🏁")
        except UnicodeEncodeError:
            typer.echo(" ")
        typer.echo("--")

    except Exception as e:  # pylint: disable=broad-except
        typer.echo(
            f"Tried to fetch MinGW-w64 from GitHub at url={MINGW64_URL} "
            f"and install in {config_default.MINGW64_COMPILER_ROOT}, but received this "
            f"Error: {e}"
        )
        sys.exit(1)


def is_mingw64_installed() -> bool:
    """Returns True if the MinGW-w64 compiler is installed."""
    if config_default.MINGW64_BIN.exists():
        return True

    return False



"""Run a dfx command"""

import sys
import subprocess
import platform
from pathlib import Path
from typing import Optional
from icpp.run_shell_cmd import run_shell_cmd


DFX = "dfx"
RUN_IN_POWERSHELL = False
if platform.win32_ver()[0]:
    DFX = "wsl --% . ~/.local/share/dfx/env; dfx"
    RUN_IN_POWERSHELL = True


def run_dfx_cmd(
    args: str,
    capture_output: bool = False,
    print_captured_output: bool = False,
    cwd: Optional[Path] = None,
    timeout_seconds: Optional[int] = None,
) -> Optional[str]:
    """Runs dfx command as a subprocess"""
    try:
        cmd = f"{DFX} {args} "
        return run_shell_cmd(
            cmd,
            capture_output=capture_output,
            print_captured_output=print_captured_output,
            cwd=cwd,
            timeout_seconds=timeout_seconds,
            run_in_powershell=RUN_IN_POWERSHELL,
        ).rstrip("\n")
    except subprocess.CalledProcessError as e:
        print(f"Failed dfx command: '{cmd}' with error: \n{e.output}")
        sys.exit(1)
    return None



"""
Set the default config variables
"""

import os
import sys
import platform
from pathlib import Path
import getpass
import icpp_candid  # pylint: disable = unused-import
from icpp import __version_wasi_sdk__, __version_rust__, __version_mingw64__

# The OS we're running on - https://stackoverflow.com/a/1857/5480536

OS_SYSTEM = platform.system()
OS_PROCESSOR = platform.processor()

#######################################################################
# Paths for logs & to install dependencies (wasi-sdk, rust, ...)
ICPP_ROOT = Path.home() / ".icpp"
ICPP_LOGS = ICPP_ROOT / "logs"

# wasi-sdk compiler
WASI_SDK_ROOT = ICPP_ROOT / "wasi-sdk"
WASI_SDK_COMPILER_ROOT = WASI_SDK_ROOT / f"{__version_wasi_sdk__}"

# mingw64 compiler (Windows only)
MINGW64_ROOT = ICPP_ROOT / "mingw64"
MINGW64_COMPILER_ROOT = MINGW64_ROOT / f"{__version_mingw64__}"
MINGW64_BIN = MINGW64_COMPILER_ROOT / "mingw64" / "bin"
if OS_SYSTEM == "Windows":
    # Add the mingw64/bin directory to the PATH
    # pylint: disable = no-member
    os.environ["PATH"] = (
        str(MINGW64_BIN)
        + os.pathsep  # pylint: disable = no-member
        + os.environ["PATH"]
    )

# Rust & dependencies
RUST_ROOT = ICPP_ROOT / "rust"
RUST_COMPILER_ROOT = RUST_ROOT / f"{__version_rust__}"
RUST_BIN = RUST_COMPILER_ROOT / "bin"
RUST_TARGET = RUST_COMPILER_ROOT / "target/wasm32-wasi/release"

if OS_SYSTEM == "Windows":
    CARGO = RUST_BIN / "cargo.exe"
    # CARGO_BINSTALL = RUST_BIN / "cargo-binstall.exe"  # Not using this yet...
    RUSTUP = RUST_BIN / "rustup.exe"
    RUSTC = RUST_BIN / "rustc.exe"
    WASI2IC = RUST_BIN / "wasi2ic.exe"
else:
    CARGO = RUST_BIN / "cargo"
    # CARGO_BINSTALL = RUST_BIN / "cargo-binstall"  # Not using this yet...
    RUSTUP = RUST_BIN / "rustup"
    RUSTC = RUST_BIN / "rustc"
    WASI2IC = RUST_BIN / "wasi2ic"

IC_WASI_POLYFILL = RUST_TARGET / "libic_wasi_polyfill.a"

# Ensure that rust is installed in the correct folder
os.environ["CARGO_TARGET_DIR"] = str(  # pylint: disable = no-member
    RUST_COMPILER_ROOT / "target"
)
os.environ["CARGO_HOME"] = str(RUST_COMPILER_ROOT)  # pylint: disable = no-member
os.environ["RUSTUP_HOME"] = str(RUST_COMPILER_ROOT)  # pylint: disable = no-member

#######################################################################
USER = getpass.getuser()

ICPP_PATH = Path(sys.modules["icpp"].__path__[0])
ICPP_DIR = str(ICPP_PATH)

ICPP_CANDID_PATH = Path(sys.modules["icpp_candid"].__path__[0])
ICPP_CANDID_DIR = str(ICPP_CANDID_PATH)

########################################################################
# icpp.toml  (can be overwritten by --config)
#
ICPP_TOML_PATH = Path("icpp.toml")

########################################################################
# The greet canister created by `icpp init`
#
ICPP_INIT_PATH = ICPP_PATH / "canisters/greet"
LOCAL_INIT_PATH = Path("./greet")

########################################################################
# build-wasm
#
IC_C_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0/*.c"))
    + list(ICPP_PATH.glob("ic/wasi_sdk_traps/*.c"))
    + list(ICPP_PATH.glob("ic/canister/*.c"))
    + list(ICPP_PATH.glob("ic/icapi/*.c"))
    + list(ICPP_CANDID_PATH.glob("candid/*.c"))
]
IC_CPP_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0/*.cpp"))
    + list(ICPP_PATH.glob("ic/canister/*.cpp"))
    + list(ICPP_PATH.glob("ic/icapi/*.cpp"))
    + list(ICPP_PATH.glob("ic/pro/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("candid/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("hooks/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("vendors/hash-library/*.cpp"))
]
IC_HEADER_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0/*.h"))
    + list(ICPP_PATH.glob("ic/wasi_sdk_traps/*.h"))
    + list(ICPP_PATH.glob("ic/canister/*.h"))
    + list(ICPP_PATH.glob("ic/icapi/*.h"))
    + list(ICPP_PATH.glob("ic/pro/*.h"))
    + list(ICPP_CANDID_PATH.glob("candid/*.h"))
    + list(ICPP_CANDID_PATH.glob("hooks/*.h"))
    + list(ICPP_CANDID_PATH.glob("vendors/hash-library/*.h"))
    + list(ICPP_CANDID_PATH.glob("vendors/cppcodec/*.hpp"))
    + list(ICPP_CANDID_PATH.glob("vendors/cppcodec/*/*.hpp"))
]

IC_C_FILES = " ".join(IC_C_FILES_LIST) + " "
IC_CPP_FILES = " ".join(IC_CPP_FILES_LIST) + " "
IC_HEADER_FILES = " ".join(IC_HEADER_FILES_LIST) + " "

SYSROOT = WASI_SDK_COMPILER_ROOT / "share/wasi-sysroot"

WASM_C_REQUIRED_FLAGS = (
    f" --target=wasm32-wasi --sysroot {SYSROOT} "
    f" -I {ICPP_DIR}/ic/canister -I {ICPP_DIR}/ic/icapi "
    f" -I {ICPP_DIR}/ic/pro -I {ICPP_DIR}/ic/ic0 "
    f" -I {ICPP_CANDID_PATH}/candid -I {ICPP_CANDID_PATH}/hooks "
    f" -I {ICPP_CANDID_PATH}/vendors "
)
WASM_CPP_REQUIRED_FLAGS = WASM_C_REQUIRED_FLAGS + " -std=c++20 "

WASM_C = WASI_SDK_COMPILER_ROOT / "bin/clang"
WASM_CPP = WASI_SDK_COMPILER_ROOT / "bin/clang++"
WASM_AR = WASI_SDK_COMPILER_ROOT / "bin/llvm-ar"
WASM_CFLAGS = (
    " -O3 -flto -fno-exceptions -fvisibility=hidden -D NDEBUG -D ICPP_VERBOSE=0 "
)
WASM_CPPFLAGS = WASM_CFLAGS
WASM_LDFLAGS = " -Wl,--lto-O3 -Wl,--strip-debug -Wl,--stack-first -Wl,--export-dynamic "
WASM_ARFLAGS = "qc"
WASM_AR_EXT = ".a"

# Utilities we can use too
CLANG_FORMAT = WASI_SDK_COMPILER_ROOT / "bin/clang-format"
CLANG_TIDY = WASI_SDK_COMPILER_ROOT / "bin/clang-tidy"
LLVM_OBJCOPY = WASI_SDK_COMPILER_ROOT / "bin/llvm-objcopy"

########################################################################
# build-native
#
MOCKIC_C_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0mock/*.c"))
    + list(ICPP_PATH.glob("ic/canister/*.c"))
    + list(ICPP_PATH.glob("ic/icapi/*.c"))
    + list(ICPP_CANDID_PATH.glob("candid/*.c"))
]

MOCKIC_CPP_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0mock/*.cpp"))
    + list(ICPP_PATH.glob("ic/canister/*.cpp"))
    + list(ICPP_PATH.glob("ic/icapi/*.cpp"))
    + list(ICPP_PATH.glob("ic/pro/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("candid/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("hooks/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("vendors/hash-library/*.cpp"))
]
MOCKIC_HEADER_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0mock/*.h"))
    + list(ICPP_PATH.glob("ic/canister/*.h"))
    + list(ICPP_PATH.glob("ic/icapi/*.h"))
    + list(ICPP_PATH.glob("ic/pro/*.h"))
    + list(ICPP_CANDID_PATH.glob("candid/*.h"))
    + list(ICPP_CANDID_PATH.glob("hooks/*.h"))
    + list(ICPP_CANDID_PATH.glob("vendors/hash-library/*.h"))
    + list(ICPP_CANDID_PATH.glob("vendors/cppcodec/*.hpp"))
    + list(ICPP_CANDID_PATH.glob("vendors/cppcodec/*/*.hpp"))
]

MOCKIC_C_FILES = " ".join(MOCKIC_C_FILES_LIST) + " "
MOCKIC_CPP_FILES = " ".join(MOCKIC_CPP_FILES_LIST) + " "
MOCKIC_HEADER_FILES = " ".join(MOCKIC_HEADER_FILES_LIST) + " "

NATIVE_COMPILER = "Clang"
NATIVE_C = "clang"
NATIVE_CPP = "clang++"
NATIVE_AR = "ar"

NATIVE_C_REQUIRED_FLAGS = (
    f" -I {ICPP_DIR}/ic/canister -I {ICPP_DIR}/ic/icapi "
    f" -I {ICPP_DIR}/ic/pro -I {ICPP_DIR}/ic/ic0mock "
    f" -I {ICPP_CANDID_PATH}/candid -I {ICPP_CANDID_PATH}/hooks "
    f" -I {ICPP_CANDID_PATH}/vendors "
)
NATIVE_CPP_REQUIRED_FLAGS = NATIVE_C_REQUIRED_FLAGS + " -std=c++20 "

NATIVE_CFLAGS = " -g -D ICPP_VERBOSE=0 "
NATIVE_CPPFLAGS = NATIVE_CFLAGS
NATIVE_LDFLAGS = " "

NATIVE_ARFLAGS = "qc"
NATIVE_AR_EXT = ".a"



"""
Defines the version for the compatible rust compiler.

(-) Release scripts might modify this automatically...
(-) do not add anything but the version number of the rust compiler here!

"""

__version__ = "1.79.0"



"""Handles 'icpp install-rust' """

import os
import sys
import platform
import subprocess
import shutil
import stat
from typing import Callable
import typer
import requests
from icpp.__main__ import app
from icpp import config_default
from icpp import __version_rust__
from icpp.run_shell_cmd import run_shell_cmd_with_log

OS_SYSTEM = platform.system()
OS_PROCESSOR = platform.processor()

LOG_FILE = config_default.ICPP_LOGS / "install_rust.log"
TIMEOUT_SECONDS = 1000


def install_rustup(nstep: int, num_steps: int) -> None:
    """Installs rustup into user's icpp folder"""
    typer.echo(f"- {nstep}/{num_steps} Installing rustup (be patient...)")

    if OS_SYSTEM == "Windows":
        # See: https://stackoverflow.com/a/73777000
        url = "https://static.rust-lang.org/rustup/dist/x86_64-pc-windows-gnu/rustup-init.exe"  # pylint: disable=line-too-long
        rustup_init_exe = config_default.RUST_ROOT / "rustup-init.exe"

        if OS_SYSTEM == "Windows":
            rustup_init_exe.unlink(missing_ok=True)

        response = requests.get(url)
        with open(rustup_init_exe, "wb") as file:
            file.write(response.content)

        cmd = (
            f"{rustup_init_exe} --no-modify-path -y "
            f'--default-toolchain="{__version_rust__}" '
            f" --default-host x86_64-pc-windows-gnu "
        )
    else:
        cmd = (
            f'curl --proto "=https" --tlsv1.2 -sSf https://sh.rustup.rs | '
            f'sh -s -- --no-modify-path -y --default-toolchain="{__version_rust__}" '
        )

    # Note: in config_default.py, we defined the environment variables for
    #       CARGO_TARGET_DIR, CARGO_HOME, RUSTUP_HOME
    #       which ensures that rust is installed in the correct folder (~/.icpp/rust)
    #
    run_shell_cmd_with_log(LOG_FILE, "w", cmd, timeout_seconds=TIMEOUT_SECONDS)

    if OS_SYSTEM == "Windows":
        rustup_init_exe.unlink(missing_ok=True)


def install_wasm32_wasi(nstep: int, num_steps: int) -> None:
    """Installs rust wasm32-wasi target into user's icpp folder"""
    typer.echo(f"- {nstep}/{num_steps} Installing wasm32-wasi target for rust compiler")
    cmd = f"{config_default.RUSTUP} target add wasm32-wasi "
    run_shell_cmd_with_log(LOG_FILE, "a", cmd, timeout_seconds=TIMEOUT_SECONDS)


def install_wasi2ic(nstep: int, num_steps: int) -> None:
    """Installs wasi2ic into user's icpp folder"""
    typer.echo(f"- {nstep}/{num_steps} Installing wasi2ic")
    cmd = (
        f"{config_default.CARGO} install "
        f"--git https://github.com/wasm-forge/wasi2ic "
        f"--tag v0.2.11 "
    )
    run_shell_cmd_with_log(LOG_FILE, "a", cmd, timeout_seconds=TIMEOUT_SECONDS)


def install_ic_wasi_polyfill(nstep: int, num_steps: int) -> None:
    """Installs ic-wasi-polyfill as a static library into user's icpp folder"""

    msg = f"- {nstep}/{num_steps} Installing ic-wasi-polyfill "
    typer.echo(msg)

    cmd = "git clone https://github.com/wasm-forge/ic-wasi-polyfill "
    run_shell_cmd_with_log(
        LOG_FILE,
        "a",
        cmd,
        cwd=config_default.RUST_COMPILER_ROOT,
        timeout_seconds=TIMEOUT_SECONDS,
    )

    cmd = "git switch --detach 2476429bb1604d3ad5a28d458b8faa61bcbe4ada "
    run_shell_cmd_with_log(
        LOG_FILE,
        "a",
        cmd,
        cwd=config_default.RUST_COMPILER_ROOT / "ic-wasi-polyfill",
        timeout_seconds=TIMEOUT_SECONDS,
    )

    cmd = f"{config_default.CARGO} build --release --target wasm32-wasi "

    #
    # The 'transient' feature use the transient file system implementation.
    # This works faster but does not take the advantage of keeping the file
    # system's state in stable memory (and the ability to keep FS state
    # between canister upgrades)
    #
    transient_memory = False
    if transient_memory:
        cmd += " --features transient "

    run_shell_cmd_with_log(
        LOG_FILE,
        "a",
        cmd,
        cwd=config_default.RUST_COMPILER_ROOT / "ic-wasi-polyfill",
        timeout_seconds=TIMEOUT_SECONDS,
    )


def remove_readonly(func: Callable[[str], None], path: str, _: None) -> None:
    """Make file writeable, then try again to apply func"""
    os.chmod(path, stat.S_IWRITE)  # pylint: disable = no-member
    func(path)


@app.command()
def install_rust() -> None:
    """Installs rust and required dependencies.

    Compiler will be installed in ~/.icpp"""

    typer.echo(f"Installing rust into {config_default.RUST_COMPILER_ROOT}")
    typer.echo(f"Details in {LOG_FILE}")

    try:
        shutil.rmtree(config_default.RUST_COMPILER_ROOT, onerror=remove_readonly)
    except FileNotFoundError:
        pass
    except OSError as e:
        typer.echo(f"Warning: {e.strerror}")

    config_default.RUST_COMPILER_ROOT.mkdir(parents=True, exist_ok=True)
    config_default.ICPP_LOGS.mkdir(parents=True, exist_ok=True)
    # ----------------------------------------------------------------

    num_steps = 4
    nstep = 1
    try:
        install_rustup(nstep, num_steps)
        nstep += 1

        install_wasm32_wasi(nstep, num_steps)
        nstep += 1

        install_wasi2ic(nstep, num_steps)
        nstep += 1

        install_ic_wasi_polyfill(nstep, num_steps)
        nstep += 1

        typer.echo("\nSuccessfully installed rust & required dependencies ")
        try:
            typer.echo("💯 🎉 🏁")
        except UnicodeEncodeError:
            typer.echo(" ")
        typer.echo("--")

    except subprocess.CalledProcessError as e:
        typer.echo(f"\nError: {e.output}")
        sys.exit(1)
    except Exception as e:  # pylint: disable=broad-except
        typer.echo(f"\nAn unexpected error occurred: {str(e)}")
        sys.exit(1)


def is_rust_installed() -> bool:
    """Returns True if rust and the required dependencies are installed."""
    required_paths = [
        config_default.RUST_COMPILER_ROOT,
        config_default.RUST_BIN,
        config_default.CARGO,
        # config_default.CARGO_BINSTALL, # Not using this yet...
        config_default.RUSTUP,
        config_default.RUSTC,
        config_default.WASI2IC,
        config_default.IC_WASI_POLYFILL,
    ]

    return all(path.exists() for path in required_paths)



"""Handles 'icpp build-library-native' """

# pylint: disable = too-many-statements
import sys
import os
import copy
import subprocess
import shutil
import concurrent.futures
from typing import Optional
from pathlib import Path
import typer
from typing_extensions import Annotated

from icpp import __version__
from icpp.__main__ import app

from icpp import config_default
from icpp.run_shell_cmd import run_shell_cmd

from icpp.decorators import requires_native_compiler
from icpp.options_build import (
    config_callback,
)

# options are: "none", "multi-threading"
CONCURRENCY = "multi-threading"


@app.command()
@requires_native_compiler()
def build_library_native(
    config: Annotated[
        str,
        typer.Option(
            help="Name of configuration toml file.",
            callback=config_callback,
        ),
    ] = "icpp.toml",
    lib_name: Annotated[
        Optional[str],
        typer.Argument(
            help="lib_name to build. If not specified, all libraries are built."
        ),
    ] = None,
) -> None:
    """Builds one or more native static debug libraries."""

    config_default.ICPP_TOML_PATH = Path(config)
    from icpp import icpp_toml  # pylint: disable = import-outside-toplevel

    typer.echo("----------------------------")
    typer.echo("Building libraries...")

    # ----------------------------------------------------------------------
    # build-library-native
    build_root_path = icpp_toml.icpp_toml_path.parent / "build-library-native"
    # typer.echo(f"Build folder: {build_root_path.resolve()}")

    if lib_name is None:
        if build_root_path.exists():
            typer.echo(f"Deleting the folder: {build_root_path} ")
            try:
                shutil.rmtree(build_root_path)
            except OSError as e:
                typer.echo(f"ERROR: During delete of {build_root_path}:\n {e.strerror}")

    if not build_root_path.exists():
        build_root_path.mkdir()

    # ----------------------------------------------------------------------
    # build-library/lib_name

    cpp_compile_flags_defaults_s = config_default.NATIVE_CPPFLAGS
    # cpp_link_flags_defaults_s = config_default.NATIVE_LDFLAGS
    c_compile_flags_defaults_s = config_default.NATIVE_CFLAGS

    def cpp_compile_file_mine(file: str, cpp_compile_cmd: str, path: Path) -> None:
        # ensure unique *.o names, even if file has same name in a subdirectory
        # pylint: disable=no-member
        output_file = (
            f"{file}".replace(f"{os.getcwd()}", "").replace("/", "_").replace("\\", "_")
        )
        cmd = f"{cpp_compile_cmd} -c {file} -o {output_file}.o"
        typer.echo(file)
        run_shell_cmd(cmd, cwd=path)

    def c_compile_file_mine(file: str, c_compile_cmd: str, path: Path) -> None:
        # ensure unique *.o names, even if file has same name in a subdirectory
        # pylint: disable=no-member
        output_file = (
            f"{file}".replace(f"{os.getcwd()}", "").replace("/", "_").replace("\\", "_")
        )
        cmd = f"{c_compile_cmd} -c {file} -o {output_file}.o"
        typer.echo(file)
        run_shell_cmd(cmd, cwd=path)

    # loop over the [[build-library]], then add routines of [[build-library-native]];
    # use flags of [[build-library-native]]
    for lib in icpp_toml.libraries:
        # when we apply the hack below, we do not want to overwrite reference data
        library = copy.deepcopy(lib)

        if (lib_name is None) or (lib_name == library["lib_name"]):
            # Find the library-native with same name, if it was provided
            library_native = None
            for lib_native in icpp_toml.libraries_native:
                if lib_native["lib_name"] == library["lib_name"]:
                    library_native = lib_native
                    break

            # hack for __ic_candid__ library. We use ONLY the native files
            if library["lib_name"] == "__ic_candid__":
                library["cpp_files_list"] = []
                library["c_files_list"] = []
                library["cpp_files"] = ""
                library["cpp_header_files"] = ""
                library["c_files"] = ""

            # create the build folder
            full_lib_name = f"{library['lib_name']}{config_default.NATIVE_AR_EXT}"

            build_path = build_root_path / library["lib_name"]
            typer.echo(f"Native library build folder: {build_path.resolve()}")

            if build_path.exists():
                # typer.echo("Deleting the build folder.")
                try:
                    shutil.rmtree(build_path)
                except OSError as e:
                    typer.echo(f"ERROR: During delete of {build_path}:\n {e.strerror}")

            if not build_path.exists():
                build_path.mkdir()

            # ----------------------------------------------------------------------

            cpp_files = library["cpp_files"]
            if library_native:
                cpp_files += library_native["cpp_files"]

            cpp_files_list = library["cpp_files_list"]
            if library_native:
                cpp_files_list += library_native["cpp_files_list"]

            cpp_include_flags = library["cpp_include_flags"]
            if library_native:
                cpp_include_flags += library_native["cpp_include_flags"]

            cpp_compile_flags_s = ""
            if library_native:
                cpp_compile_flags_s += library_native["cpp_compile_flags_s"]

            c_files = library["c_files"]
            if library_native:
                c_files += library_native["c_files"]

            c_files_list = library["c_files_list"]
            if library_native:
                c_files_list += library_native["c_files_list"]

            c_include_flags = library["c_include_flags"]
            if library_native:
                c_include_flags += library_native["c_include_flags"]

            c_compile_flags_s = ""
            if library_native:
                c_compile_flags_s += library_native["c_compile_flags_s"]

            if library_native:
                if library_native["overwrite_default_CPPFLAGS"]:
                    cpp_compile_flags_defaults_s = library_native[
                        "cpp_compile_flags_defaults_s"
                    ]
                if library_native["overwrite_default_CFLAGS"]:
                    c_compile_flags_defaults_s = library_native[
                        "c_compile_flags_defaults_s"
                    ]

            cpp_compile_cmd_mine = (
                f"{config_default.NATIVE_CPP} "
                f"{config_default.NATIVE_CPP_REQUIRED_FLAGS} "
                f"{cpp_compile_flags_defaults_s} "
                f"{cpp_include_flags} "
                f"{cpp_compile_flags_s} "
            )

            c_compile_cmd_mine = (
                f"{config_default.NATIVE_C} "
                f"{config_default.NATIVE_C_REQUIRED_FLAGS} "
                f"{c_compile_flags_defaults_s} "
                f"{c_include_flags} {c_compile_flags_s} "
            )

            try:
                # ----------------------------------------------------------------------
                # compile C++ files, if we have any
                if len(cpp_files.strip()) > 0:
                    if CONCURRENCY == "multi-threading":
                        typer.echo("--")
                        typer.echo(
                            f"Compiling C++ files for library: {library['lib_name']}"
                        )
                        typer.echo(f"Compile command: {cpp_compile_cmd_mine}")
                        with concurrent.futures.ThreadPoolExecutor() as executor:
                            executor.map(
                                lambda file, cpp_compile_cmd=cpp_compile_cmd_mine, build_path=build_path: cpp_compile_file_mine(  # pylint: disable=line-too-long
                                    file, cpp_compile_cmd, build_path
                                ),
                                cpp_files_list,
                            )
                    else:
                        cmd = f"{cpp_compile_cmd_mine} -c {cpp_files}"

                        typer.echo("--")
                        typer.echo(
                            f"Compiling C++ files for library: {library['lib_name']}"
                        )
                        typer.echo(cmd)
                        run_shell_cmd(cmd, cwd=build_path)

                # ----------------------------------------------------------------------
                # compile C files, if we have any
                if len(c_files.strip()) > 0:
                    if CONCURRENCY == "multi-threading":
                        typer.echo("--")
                        typer.echo(
                            f"Compiling C files for library: {library['lib_name']}"
                        )
                        typer.echo(f"Compile command: {c_compile_cmd_mine}")
                        with concurrent.futures.ThreadPoolExecutor() as executor:
                            executor.map(
                                lambda file, c_compile_cmd=c_compile_cmd_mine, build_path=build_path: c_compile_file_mine(  # pylint: disable=line-too-long
                                    file, c_compile_cmd, build_path
                                ),
                                c_files_list,
                            )
                    else:
                        cmd = f"{c_compile_cmd_mine} -c {c_files}"

                        typer.echo("--")
                        typer.echo(
                            f"Compiling C files for library: {library['lib_name']}"
                        )
                        typer.echo(cmd)
                        run_shell_cmd(cmd, cwd=build_path)

                # ----------------------------------------------------------------------
                # create the library
                cmd = (
                    f"{config_default.NATIVE_AR} "
                    f"{config_default.NATIVE_ARFLAGS} "
                    f"{full_lib_name} *.o"
                )

                typer.echo("--")
                typer.echo(f"Creating the static library: {full_lib_name}")
                typer.echo(cmd)
                run_shell_cmd(cmd, cwd=build_path)

            except subprocess.CalledProcessError as e:
                sys.exit(e.returncode)
            except Exception as e:  # pylint: disable=broad-except
                typer.echo(
                    f"ERROR: something went wrong in module "
                    f"'commands_build_library_native.py' \n "
                    f"Exception is: \n{e}"
                )
                sys.exit(1)

            # ----------------------------------------------------------------------
            typer.echo("--")
            typer.echo("All done building the native static library:")
            typer.echo(f"{build_path.resolve()}/{full_lib_name}")
            try:
                typer.echo("✔️")
            except UnicodeEncodeError:
                typer.echo(" ")

            # ----------------------------------------------------------------------
            typer.echo("-----")
